#include <dos.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "define.h"
#include "mode15k.h"
#include "port.h"


extern SCRLIST ScrModeList[];


int GetMode15KHz(CRTCPARAM *cp, int xres, int yres)
{
        SCRLIST *ListPtr;
        int idx;
        char found;

        ListPtr = ScrModeList;

        idx = 0;
        found=0;

        while(1)
        {
                if( (ListPtr + idx)->forx == 0 || (ListPtr + idx)->fory == 0)
                {
                        /* Mode not found/force 640x480 then */
                        found=1;
                        break;
                }
                else if( (ListPtr + idx)->forx == xres && (ListPtr + idx)->fory == yres)
                {
                        found=1;
                        break;
                }
                idx += 1;
        }

        if (found)
        {
                *cp = *((ListPtr + idx)->cp);
        }

        return found;
}


void SetGenericVGARegister(CRTCPARAM *cp, int h_sync_delay, int h_disp_delay)
{
        int d0;
        int d1;
        int HDispSkew;
        int HSyncDelay;

        HSyncDelay = h_sync_delay & 0x03;
        HDispSkew  = h_disp_delay & 0x03;

        // Unlock
        SetCrtcReg(0x11, BitMov(GetCrtcReg(0x11), 7, 0, 0));

        // HTotal
        d0 = cp->HTotal;
        d0 = (d0 / 8) - 5;
        SetCrtcReg(0x00, d0 & 0xFF);

        // HDisp
        d0 = cp->HDisp;
        d0 = (d0 / 8) - 1;
        SetCrtcReg(0x01, d0 & 0xFF);

        // HBlankStart
        d0 = cp->HBStart;
        d0 = d0 / 8;
        SetCrtcReg(0x02, d0 & 0xFF);

        // HBlankEnd
        d0 = cp->HBEnd;
        d0 = d0 / 8;
        d1 = GetCrtcReg(0x03);
        d1 = BitMov(d1, 0, d0, 0);
        d1 = BitMov(d1, 1, d0, 1);
        d1 = BitMov(d1, 2, d0, 2);
        d1 = BitMov(d1, 3, d0, 3);
        d1 = BitMov(d1, 4, d0, 4);
        d1 = BitMov(d1, 5, HDispSkew, 0);       // horizontal disp skew
        d1 = BitMov(d1, 6, HDispSkew, 1);       // horizontal disp skew
        d1 = BitMov(d1, 7, 1, 0);                       // compatible read
        SetCrtcReg(0x03, d1);
        SetCrtcReg(0x05, BitMov(GetCrtcReg(0x05), 7, d0, 5));

        // HSyncStart
        d0 = cp->HSStart;
        d0 = d0 / 8;
        SetCrtcReg(0x04, d0 & 0xFF);

        // HSyncEnd
        d0 = cp->HSEnd;
        d0 = d0 / 8;
        d1 = GetCrtcReg(0x05);
        d1 = BitMov(d1, 0, d0, 0);
        d1 = BitMov(d1, 1, d0, 1);
        d1 = BitMov(d1, 2, d0, 2);
        d1 = BitMov(d1, 3, d0, 3);
        d1 = BitMov(d1, 4, d0, 4);
        d1 = BitMov(d1, 5, HSyncDelay, 0);
        d1 = BitMov(d1, 6, HSyncDelay, 1);
        SetCrtcReg(0x05, d1);

        // VTotal
        d0 = cp->VTotal;
        d0 = d0 - 2;
        SetCrtcReg(0x06, d0 & 0xFF);
        SetCrtcReg(0x07, BitMov(GetCrtcReg(0x07), 0, d0, 8));
        SetCrtcReg(0x07, BitMov(GetCrtcReg(0x07), 5, d0, 9));

        // VDisp
        d0 = cp->VDisp;
        d0 = d0 - 1;
        SetCrtcReg(0x12, d0 & 0xFF);
        SetCrtcReg(0x07, BitMov(GetCrtcReg(0x07), 1, d0, 8));
        SetCrtcReg(0x07, BitMov(GetCrtcReg(0x07), 6, d0, 9));


        // VBlankStart
        d0 = cp->VBStart;
        SetCrtcReg(0x15, d0 & 0xFF);
        SetCrtcReg(0x07, BitMov(GetCrtcReg(0x07), 3, d0, 8));
        SetCrtcReg(0x09, BitMov(GetCrtcReg(0x09), 5, d0, 9));

        // VBlankEnd
        d0 = cp->VBEnd;
        SetCrtcReg(0x16, d0 & 0xFF);

        // VSyncStart
        d0 = cp->VSStart;
        SetCrtcReg(0x10, d0 & 0xFF);
        d1 = GetCrtcReg(0x07);
        d1 = BitMov(d1, 2, d0,  8);  // bit 8
        d1 = BitMov(d1, 7, d0,  9);  // bit 9
        SetCrtcReg(0x07, d1);

        // VSyncEnd
        d0 = cp->VSEnd;
        d0 = d0 & 0x0F;
        d1 = GetCrtcReg(0x11) & 0xF0;
        d1 = d1 | d0;
        SetCrtcReg(0x11, d1);

        // Double Scan Dis.
        SetCrtcReg(0x09, BitMov(GetCrtcReg(0x09), 7, 0, 0));

}

void AdjustCenter(CRTCPARAM *cp, int center_x, int center_y)
{
    int d0;

    if(center_x < 0){
        if((cp->HSStart + center_x) < (cp->HDisp + 8)){
            d0 = cp->HSStart;
            cp->HSStart = cp->HDisp + 8;
            cp->HSEnd  -= (d0 - cp->HSStart);
        }else{
            cp->HSStart = cp->HSStart + center_x;
            cp->HSEnd   = cp->HSEnd   + center_x;
        }
    }else{
        if((cp->HSEnd + center_x) > (cp->HTotal - 16)){
            d0 = cp->HSEnd;
            cp->HSEnd    = cp->HTotal - 16;
            cp->HSStart += (cp->HSEnd - d0);
        }else{
            cp->HSEnd   = cp->HSEnd   + center_x;
            cp->HSStart = cp->HSStart + center_x;
        }
    }

    if(center_y < 0){
        if((cp->VSStart + center_y) < (cp->VDisp + 1)){
            d0 = cp->VSStart;
            cp->VSStart = cp->VDisp + 1;
            cp->VSEnd  -= (d0 - cp->VSStart);
        }else{
            cp->VSStart = cp->VSStart + center_y;
            cp->VSEnd   = cp->VSEnd   + center_y;
        }
    }else{
        if((cp->VSEnd + center_y) > (cp->VTotal - 2)){
            d0 = cp->VSEnd;
            cp->VSEnd    = cp->VTotal - 2;
            cp->VSStart += (cp->VSEnd - d0);
        }else{
            cp->VSEnd   = cp->VSEnd   + center_y;
            cp->VSStart = cp->VSStart + center_y;
        }
    }
}




long CalcDotclock(long Vert001Hz, int HTotal, int VTotal, int interlace)
{
        double f0;

        f0 = (double)VTotal;
        if(interlace != 0){
                f0 += 0.5;
        }
        f0 = f0 * (double)HTotal;
        f0 = f0 * (double)Vert001Hz;
        f0 = f0 / 100.0;

        return (long)f0;
}

long CalcHorizontalHz(long dotclockHz, int HTotal)
{
        return dotclockHz / HTotal;
}

long CalcVert001Hz(long dotclockHz, int HTotal, int VTotal, int interlace)
{
        double f0;

        f0 = (double)VTotal;
        if(interlace != 0){
                f0 += 0.5;
        }
        f0 = f0 * (double)HTotal;
        f0 = (double)dotclockHz / f0;
        f0 = f0 * 100.0;

        return (long)f0;
}

void SetHVPolarity(int HPolar, int VPolar)
{
        BYTE d0;

        if(HPolar != 0) HPolar = 1;
        if(VPolar != 0) VPolar = 1;

        d0 = _inb(0x3CC);
        d0 = d0 & 0x3F;
        d0 = d0 | (VPolar << 7) | (HPolar << 6);
        _outb(0x3C2, d0);
}

DWORD BitMov(DWORD dstdata, int dstbitpos, DWORD srcdata, int srcbitpos)
{
        DWORD d0;
        DWORD d1;

        d1 = srcdata >> srcbitpos;
        d1 = d1 & 1;
        d0 = dstdata & (~(1 << dstbitpos));
        d1 = d1 << dstbitpos;
        d0 = d0 | d1;

        return d0;
}

void SetSeqReg(BYTE __idx, BYTE data)
{
        _outb(0x3c4, __idx);
        _outb(0x3c5, data);
}

BYTE GetSeqReg(BYTE __idx)
{
        BYTE d0;

        _outb(0x3c4, __idx);
        d0 = _inb(0x3c5);
        return d0;
}

BYTE GetAtrbData(BYTE __idx)
{
        _inb(0x3da);
        _outb(0x3c0, __idx | (1 << 5));
        return _inb(0x3c1);
}

void SetAtrbData(BYTE __idx, BYTE data)
{
        _inb(0x3da);
        _outb(0x3c0, __idx | 0x20);
        _outb(0x3c0, data);
}

int GetCrtcAddress(void)
{
    if((_inb(0x3CC) & 0x01) == 0)
        return 0x3B4;
    else
        return 0x3D4;
}

void SetCrtcReg(BYTE __idx, BYTE data)
{
    _outb(GetCrtcAddress(),    __idx);
    _outb(GetCrtcAddress() + 1, data);
}

BYTE GetCrtcReg(BYTE __idx)
{
    BYTE d0;

    _outb(GetCrtcAddress(), __idx);
    d0 = _inb(GetCrtcAddress() + 1);
    return d0;
}


int GetSeqDivider(void)
{
        int d0;

        d0      = (GetSeqReg(0x01) & 0x08) >>3;
        d0      = d0 + 1;
        return d0;
}

void _outb(unsigned int edx, BYTE al)
{
        outportb(edx, al);
}

BYTE _inb(unsigned int edx)
{
        return inportb(edx);
}

void _outw(unsigned int edx, WORD ax)
{
        outport(edx, ax);
}
WORD _inw(unsigned int edx)
{
        return inport(edx);
}

void _outl(unsigned int edx, DWORD eax)
{
        outportl(edx, eax);
}

DWORD _inl(unsigned int edx)
{
        return inportl(edx);
}

void DisableGenerateSignal(void)
{
        int d0;

        d0 = GetCrtcReg(0x17);
        d0 = d0 & 0x7F;
        SetCrtcReg(0x17, d0);

}

void EnableGenerateSignal(void)
{
        BYTE d0;

        d0 = GetCrtcReg(0x17);
        d0 = d0 | 0x80;
        SetCrtcReg(0x17, d0);
}

void SetSeqDiv(int div)
{
    if(div == 2)
        div = 1;
    else
        div = 0;
    SetSeqReg(0x01, BitMov(GetSeqReg(0x01), 3, div, 0));
}

void Set25MHz(void)
{
    BYTE d0;

    d0 = _inb(0x3CC);
    d0 = BitMov(d0, 2, 0, 0);
    d0 = BitMov(d0, 3, 0, 0);
    _outb(0x3C2, d0);               /*  25MHz (640 mode) */
    SetSeqDiv(1);
}

void Set28MHz(void)
{
    BYTE d0;

    d0 = _inb(0x3CC);
    d0 = BitMov(d0, 2, 1, 0);
    d0 = BitMov(d0, 3, 0, 0);
    _outb(0x3C2, d0);               /* 28MHz (720 mode) */
    SetSeqDiv(1);
}

void Set12MHz(void)
{
    BYTE d0;

    d0 = _inb(0x3CC);
    d0 = BitMov(d0, 2, 0, 0);
    d0 = BitMov(d0, 3, 0, 0);
    _outb(0x3C2, d0);                                   /* 25MHz (640 mode) */
    SetSeqDiv(2);
}

void Set14MHz(void)
{
    BYTE d0;

    d0 = _inb(0x3CC);
    d0 = BitMov(d0, 2, 1, 0);
    d0 = BitMov(d0, 3, 0, 0);
    _outb(0x3C2, d0);                                   /* 28MHz (720 mode) */
    SetSeqDiv(2);
}


void SetTextHeight(int height)
{
    BYTE d0;

    height = (height - 1) & 0x1F;
    d0  = GetCrtcReg(0x09) & ~0x1F;
    d0 |= height;
    SetCrtcReg(0x09, d0);
}

void SetTextWidth(int width)
{
    int flag;

    if(width == 8)
        flag = 1;
    else
        flag = 0;
    SetSeqReg(0x01, (GetSeqReg(0x01) & 0xFE) | flag);
}

BYTE GetVGABytesPerLine(void)
{
    return GetCrtcReg(0x13);
}

void SetVGABytesPerLine(BYTE d0)
{
    SetCrtcReg(0x13, d0);
}


static void SetDoubleScan(int flag)
{
    if(flag == 0){
        SetCrtcReg(0x09, BitMov(GetCrtcReg(0x09), 7, 0, 0));
    }else{
        SetCrtcReg(0x09, BitMov(GetCrtcReg(0x09), 7, 1, 0));
    }
}


/* can restore dosbox <-> full screen */
void SetText80x25(int center_x, int center_y, int h_sync_delay, int h_disp_delay)
{
    CRTCPARAM cp;

    GetMode15KHz(&cp, 640, 200);     /* 80x25 */
    AdjustCenter(&cp, center_x, center_y);

    DisableGenerateSignal();
    Set14MHz();
    SetGenericVGARegister(&cp, h_sync_delay, h_disp_delay);
    SetDoubleScan(cp.doublescan);
    SetHVPolarity(cp.hpolarity, cp.vpolarity);
    SetTextHeight(8);
    SetTextWidth(8);
    EnableGenerateSignal();
}
